@@ -4,7 +4,6 @@ import inspect
 import itertools
 import logging
 from typing import TYPE_CHECKING
-import weakref
 
 from IPython import get_ipython
 try:
@@ -12,9 +11,9 @@ try:
 except ImportError:
     pandas = None
 
-from nbsafety.analysis import AttrSubSymbolChain, CallPoint
+from nbsafety.analysis.attr_symbols import AttrSubSymbolChain, CallPoint
 from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
-from nbsafety.singletons import nbs, nbs_check_init
+from nbsafety.singletons import nbs
 
 if TYPE_CHECKING:
     from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
@@ -65,11 +64,11 @@ class Scope:
             return self.parent_scope.non_namespace_parent_scope
         return self.parent_scope
 
-    def make_child_scope(self, scope_name, obj_id=None) -> Scope:
-        if obj_id is None:
+    def make_child_scope(self, scope_name, obj=None) -> Scope:
+        if obj is None:
             return Scope(scope_name, parent_scope=self)
         else:
-            return NamespaceScope(obj_id, scope_name, parent_scope=self)
+            return NamespaceScope(obj, scope_name, parent_scope=self)
 
     def put(self, name: SupportedIndexType, val: DataSymbol):
         self._data_symbol_by_name[name] = val
@@ -124,7 +123,7 @@ class Scope:
             if dsym is not None and next_dsym is None:
                 # HUGE HACK: prevents us from checking namespace symbols unless entire namespace is stale
                 # TODO: get rid of this check once namespace symbols created for dictionary literals
-                if dsym.is_stale and dsym.defined_cell_num >= dsym.required_cell_num:
+                if dsym.is_stale and dsym.timestamp >= dsym.required_timestamp:
                     dsym = None
                 break
             dsym, next_dsym = next_dsym, None
@@ -172,7 +171,7 @@ class Scope:
         self,
         name: SupportedIndexType,
         obj: Any,
-        deps: Set[DataSymbol],
+        deps: Iterable[DataSymbol],
         stmt_node: ast.AST,
         overwrite: bool = True,
         is_subscript: bool = False,
@@ -192,15 +191,18 @@ class Scope:
             is_anonymous=is_anonymous,
             class_scope=class_scope
         )
-        dsym, old_dsym, old_id = self._upsert_data_symbol_for_name_inner(
+        deps = set(deps)  # make a copy since we mutate it (see below fixme)
+        dsym, prev_dsym, prev_obj = self._upsert_data_symbol_for_name_inner(
             name,
             obj,
-            deps,
+            deps,  # FIXME: this updates deps, which is super super hacky
             symbol_type,
             stmt_node,
             implicit=implicit,
         )
-        dsym.update_deps(deps, overwrite=overwrite, propagate=propagate)
+        dsym.update_deps(
+            deps, prev_obj=prev_obj, overwrite=overwrite, propagate=propagate, refresh=not implicit
+        )
         return dsym
 
     def _upsert_data_symbol_for_name_inner(
@@ -211,23 +213,23 @@ class Scope:
         symbol_type: DataSymbolType,
         stmt_node: ast.AST,
         implicit: bool = False,
-    ) -> Tuple[DataSymbol, Optional[DataSymbol], Optional[int]]:
-        old_id = None
-        old_dsym = self.lookup_data_symbol_by_name_this_indentation(
+    ) -> Tuple[DataSymbol, Optional[DataSymbol], Optional[Any]]:
+        prev_obj = None
+        prev_dsym = self.lookup_data_symbol_by_name_this_indentation(
             name, is_subscript=symbol_type == DataSymbolType.SUBSCRIPT, skip_cloned_lookup=True,
         )
         if implicit and symbol_type != DataSymbolType.ANONYMOUS:
-            assert old_dsym is None, 'expected None, got %s' % old_dsym
-        if old_dsym is not None and self.is_globally_accessible:
-            old_id = old_dsym.cached_obj_id
+            assert prev_dsym is None, 'expected None, got %s' % prev_dsym
+        if prev_dsym is not None and self.is_globally_accessible:
+            prev_obj = DataSymbol.NULL if prev_dsym.obj is None else prev_dsym.obj
             # TODO: handle case where new dc is of different type
-            if name in self.data_symbol_by_name(old_dsym.is_subscript) and old_dsym.symbol_type == symbol_type:
-                old_dsym.update_obj_ref(obj, refresh_cached=False)
+            if name in self.data_symbol_by_name(prev_dsym.is_subscript) and prev_dsym.symbol_type == symbol_type:
+                prev_dsym.update_obj_ref(obj, refresh_cached=False)
                 # old_dsym.update_type(symbol_type)
                 # if we're updating a pre-existing one, it should not be an implicit upsert
                 assert stmt_node is not None
-                old_dsym.update_stmt_node(stmt_node)
-                return old_dsym, old_dsym, old_id
+                prev_dsym.update_stmt_node(stmt_node)
+                return prev_dsym, prev_dsym, prev_obj
             else:
                 # In this case, we are copying from a class and we need the dsym from which we are copying
                 # as able to propagate to the new dsym.
@@ -247,17 +249,17 @@ class Scope:
                 # If we do this, then we should go back to explicitly adding the dep as follows:
                 # EDIT: added check to avoid propagating along class -> instance edge when class not redefined, so now
                 # it is important to explicitly add this dep.
-                deps.add(old_dsym)
+                deps.add(prev_dsym)
         if isinstance(self, NamespaceScope) and symbol_type == DataSymbolType.DEFAULT and self.cloned_from is not None:
             # add the cloned symbol as a dependency of the symbol about to b ecreated
             new_dep = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name, is_subscript=False)
             if new_dep is not None:
                 deps.add(new_dep)
         dsym = DataSymbol(
-            name, symbol_type, obj, self, stmt_node=stmt_node, parents=deps, refresh_cached_obj=False, implicit=implicit
+            name, symbol_type, obj, self, stmt_node=stmt_node, refresh_cached_obj=False, implicit=implicit
         )
         self.put(name, dsym)
-        return dsym, old_dsym, old_id
+        return dsym, prev_dsym, prev_obj
 
     def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
         dsym = self._data_symbol_by_name.pop(name, None)
@@ -268,10 +270,6 @@ class Scope:
     def is_global(self):
         return self.parent_scope is None
 
-    @property
-    def is_garbage(self):
-        return False
-
     @property
     def is_globally_accessible(self):
         return self.is_global or (self.is_namespace_scope and self.parent_scope.is_globally_accessible)
@@ -306,8 +304,8 @@ class Scope:
         else:
             return self.scope_name
 
-    def make_namespace_qualified_name(self, dc: DataSymbol) -> str:
-        return str(dc.name)
+    def make_namespace_qualified_name(self, dsym: DataSymbol) -> str:
+        return str(dsym.name)
 
 
 class NamespaceScope(Scope):
@@ -319,45 +317,57 @@ class NamespaceScope(Scope):
         super().__init__(*args, **kwargs)
         self.cloned_from: Optional[NamespaceScope] = None
         self.child_clones: List[NamespaceScope] = []
-        obj_ref, obj_id = self._update_obj_ref_inner(obj)
+        self.obj = obj
+        self.cached_obj_id = id(obj)
+        if obj is not None and not isinstance(obj, int) and id(obj) in nbs().namespaces:  # pragma: no cover
+            msg = 'namespace already registered for %s' % obj
+            if nbs().is_develop:
+                raise ValueError(msg)
+            else:
+                logger.warning(msg)
+        nbs().namespaces[id(obj)] = self
         self._tombstone = False
-        self._obj_ref = obj_ref
-        self.obj_id = obj_id
-        nbs().namespaces[obj_id] = self
-        self.max_defined_timestamp = 0
+        # this timestamp needs to be bumped in DataSymbol refresh()
+        self.max_descendent_timestamp = -1
         self._subscript_data_symbol_by_name: Dict[SupportedIndexType, DataSymbol] = {}
+        self.namespace_stale_symbols: Set[DataSymbol] = set()
 
     def __bool__(self):
         # in order to override if __len__ returns 0
         return True
 
     def __len__(self):
-        obj = self._obj_ref()
-        if not isinstance(obj, (dict, list, tuple)):
-            raise TypeError("tried to get length of non-container namespace %s: %s", self, obj)
-        return len(obj)
+        if not isinstance(self.obj, (dict, list, tuple)):  # pragma: no cover
+            raise TypeError("tried to get length of non-container namespace %s: %s", self, self.obj)
+        return len(self.obj)
 
-    def _iter_inner(self, obj):
-        for i in range(len(obj)):
+    def _iter_inner(self):
+        for i in range(len(self.obj)):
             yield self.lookup_data_symbol_by_name_this_indentation(i, is_subscript=True)
 
     def __iter__(self):
-        obj = self._obj_ref()
-        if not isinstance(obj, (list, tuple)):
-            raise TypeError("tried to iterate through non-sequence namespace %s: %s", self, obj)
+        if not isinstance(self.obj, (list, tuple)):  # pragma: no cover
+            raise TypeError("tried to iterate through non-sequence namespace %s: %s", self, self.obj)
         # do the validation before starting the generator part so that we raise immediately
-        return self._iter_inner(obj)
+        return self._iter_inner()
 
-    def _items_inner(self, obj):
-        for key in obj.keys():
+    def _items_inner(self):
+        for key in self.obj.keys():
             yield key, self.lookup_data_symbol_by_name_this_indentation(key, is_subscript=True)
 
     def items(self):
-        obj = self._obj_ref()
-        if not isinstance(obj, dict):
-            raise TypeError("tried to get iterate through items of non-dict namespace: %s", obj)
+        if not isinstance(self.obj, dict):  # pragma: no cover
+            raise TypeError("tried to get iterate through items of non-dict namespace: %s", self.obj)
         # do the validation before starting the generator part so that we raise immediately
-        return self._items_inner(obj)
+        return self._items_inner()
+
+    @property
+    def obj_id(self):
+        return self.cached_obj_id
+
+    @property
+    def is_anonymous(self):
+        return self.scope_name == NamespaceScope.ANONYMOUS
 
     @property
     def is_garbage(self):
@@ -371,37 +381,12 @@ class NamespaceScope(Scope):
         else:
             return dsym.is_subscript
 
-    def get_obj(self) -> Any:
-        return self._obj_ref()
-
     def update_obj_ref(self, obj):
-        obj_ref, obj_id = self._update_obj_ref_inner(obj)
         self._tombstone = False
-        self._obj_ref = obj_ref
-        self.obj_id = obj_id
-        nbs().namespaces[obj_id] = self
-
-    def _update_obj_ref_inner(self, obj):
-        try:
-            obj_ref = weakref.ref(obj, self._obj_reference_expired_callback)
-        except TypeError:
-            obj_ref = lambda: obj
-        obj_id = id(obj)
-        return obj_ref, obj_id
-
-    def clear_namespace(self, prev_obj_id):
-        if prev_obj_id != self.obj_id and prev_obj_id in nbs().namespaces:
-            raise ValueError('precondition failed; namespace should no longer be registered before we can clear')
-        self._data_symbol_by_name.clear()
-        self._subscript_data_symbol_by_name.clear()
-
-    def _obj_reference_expired_callback(self, *_):
-        self._tombstone = True
-        safety = nbs_check_init()
-        if safety is None:
-            # can happen e.g. if program is exiting
-            return
-        safety.garbage_namespace_obj_ids.add(self.obj_id)
+        nbs().namespaces.pop(self.cached_obj_id, None)
+        self.obj = obj
+        self.cached_obj_id = id(obj)
+        nbs().namespaces[self.cached_obj_id] = self
 
     def data_symbol_by_name(self, is_subscript=False):
         if is_subscript:
@@ -410,23 +395,19 @@ class NamespaceScope(Scope):
             return self._data_symbol_by_name
 
     def clone(self, obj: Any):
-        cloned = NamespaceScope(obj, nbs())
-        cloned.__dict__ = dict(self.__dict__)
+        cloned = NamespaceScope(obj, self.scope_name, self.parent_scope)
         cloned.cloned_from = self
-        cloned.update_obj_ref(obj)
-        cloned._data_symbol_by_name = {}
-        cloned._subscript_data_symbol_by_name = {}
         self.child_clones.append(cloned)
         return cloned
 
     def fresh_copy(self, obj: Any):
         return NamespaceScope(obj, self.scope_name, self.parent_scope)
 
-    def make_namespace_qualified_name(self, dc: DataSymbol):
+    def make_namespace_qualified_name(self, dsym: DataSymbol):
         path = self.full_namespace_path
-        name = str(dc.name)
+        name = str(dsym.name)
         if path:
-            if dc.is_subscript:
+            if dsym.is_subscript:
                 return f'{path}[{name}]'
             else:
                 return f'{path}.{name}'
@@ -444,17 +425,16 @@ class NamespaceScope(Scope):
         else:
             ret = self._data_symbol_by_name.get(name, None)
         if not skip_cloned_lookup and ret is None and self.cloned_from is not None and not is_subscript and isinstance(name, str):
-            if name not in getattr(self._obj_ref(), '__dict__', {}):
+            if name not in getattr(self.obj, '__dict__', {}):
                 # only fall back to the class sym if it's not present in the corresponding obj for this scope
                 ret = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name, is_subscript=is_subscript)
         return ret
 
     def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
-        logger.info("delete %s from %s", name, self)
         if is_subscript:
             dsym = self._subscript_data_symbol_by_name.pop(name, None)
-            if dsym is None and name == -1 and isinstance(self._obj_ref(), list):
-                name = len(self._obj_ref())  # it will have already been deleted, so don't subtract 1
+            if dsym is None and name == -1 and isinstance(self.obj, list):
+                name = len(self.obj)  # it will have already been deleted, so don't subtract 1
                 dsym = self._subscript_data_symbol_by_name.pop(name, None)
             if dsym is not None:
                 dsym.update_deps(set(), deleted=True)
@@ -474,29 +454,17 @@ class NamespaceScope(Scope):
             dsym_collections_to_chain.append(self.cloned_from.all_data_symbols_this_indentation())
         return itertools.chain(*dsym_collections_to_chain)
 
-    @property
-    def num_subscript_symbols(self):
-        return len(self._subscript_data_symbol_by_name)
-
-    @property
-    def num_dotted_symbols(self):
-        return len(self._data_symbol_by_name)
-
-    @property
-    def num_symbols(self):
-        return self.num_dotted_symbols + self.num_subscript_symbols
-
     def put(self, name: SupportedIndexType, val: DataSymbol):
         if val.is_subscript:
             self._subscript_data_symbol_by_name[name] = val
+        elif not isinstance(name, str):  # pragma: no cover
+            raise TypeError('%s should be a string' % name)
         else:
-            if not isinstance(name, str):
-                raise TypeError('%s should be a string' % name)
             self._data_symbol_by_name[name] = val
         val.containing_scope = self
 
     def refresh(self):
-        self.max_defined_timestamp = nbs().cell_counter()
+        self.max_descendent_timestamp = nbs().cell_counter()
 
     def get_earliest_ancestor_containing(self, obj_id: int, is_subscript: bool) -> Optional[NamespaceScope]:
         # TODO: test this properly
@@ -505,8 +473,7 @@ class NamespaceScope(Scope):
             ret = self.namespace_parent_scope.get_earliest_ancestor_containing(obj_id, is_subscript)
         if ret is not None:
             return ret
-        set_to_check = map(lambda dsym: dsym.obj_id, self.all_data_symbols_this_indentation(is_subscript=is_subscript))
-        if obj_id in set_to_check:
+        if obj_id in (dsym.obj_id for dsym in self.all_data_symbols_this_indentation(is_subscript=is_subscript)):
             return self
         else:
             return None
