@@ -5,6 +5,7 @@ import logging
 from typing import cast, TYPE_CHECKING
 import sys
 
+from nbsafety.analysis.attr_symbols import resolve_slice_to_constant
 from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
 from nbsafety.utils import fast
 
@@ -20,9 +21,11 @@ def _maybe_convert_ast_subscript(subscript: ast.AST) -> ast.AST:
     if isinstance(subscript, ast.Index):
         return subscript.value  # type: ignore
     elif isinstance(subscript, ast.Slice):
+        lower = fast.NameConstant(None) if subscript.lower is None else subscript.lower
+        upper = fast.NameConstant(None) if subscript.upper is None else subscript.upper
         return fast.Call(
             func=fast.Name('slice', ast.Load()),
-            args=[subscript.lower, subscript.upper] + ([] if subscript.step is None else [subscript.step]),
+            args=[lower, upper] + ([] if subscript.step is None else [subscript.step]),
             keywords=[],
         )
     else:
@@ -44,11 +47,17 @@ class AstEavesdropper(ast.NodeTransformer):
 
     def _make_tuple_event_for(self, node: ast.AST, event: TraceEvent, orig_node_id=None, **kwargs):
         with fast.location_of(node):
-            tuple_node = fast.Tuple([fast.Call(
-                func=self._emitter_ast(),
-                args=[event.to_ast(), self._get_copy_id_ast(orig_node_id or node)],
-                keywords=[] if len(kwargs) == 0 else fast.kwargs(**kwargs),
-            ), node], ast.Load())
+            tuple_node = fast.Tuple(
+                [
+                    fast.Call(
+                        func=self._emitter_ast(),
+                        args=[event.to_ast(), self._get_copy_id_ast(orig_node_id or node)],
+                        keywords=[] if len(kwargs) == 0 else fast.kwargs(**kwargs),
+                    ),
+                    node,
+                ],
+                ast.Load()
+            )
             slc: Union[ast.Constant, ast.Num, ast.Index] = fast.Num(1)
             if sys.version_info < (3, 9):
                 slc = fast.Index(slc)
@@ -88,7 +97,7 @@ class AstEavesdropper(ast.NodeTransformer):
             if isinstance(attr_or_sub, (ast.Slice, ast.ExtSlice)):
                 elts = attr_or_sub.elts if isinstance(attr_or_sub, ast.Tuple) else attr_or_sub.dims  # type: ignore
                 elts = [_maybe_convert_ast_subscript(elt) for elt in elts]
-                attr_or_sub = fast.Tuple(elts)
+                attr_or_sub = fast.Tuple(elts, ast.Load())
         return self.visit_Attribute_or_Subscript(node, cast(ast.expr, attr_or_sub), call_context=call_context)
 
     def _maybe_wrap_symbol_in_before_after_tracing(
@@ -160,6 +169,13 @@ class AstEavesdropper(ast.NodeTransformer):
             if isinstance(node.value, ast.Name):
                 extra_args = fast.kwargs(obj_name=fast.Str(node.value.id))
 
+            subscript_name = None
+            if isinstance(node, ast.Subscript):
+                slice_val = resolve_slice_to_constant(node)
+                if isinstance(slice_val, ast.Name):
+                    # TODO: this should be more general than just simple ast.Name subscripts
+                    subscript_name = slice_val.id
+
             with self.attrsub_context(node):
                 node.value = fast.Call(
                     func=self._emitter_ast(),
@@ -172,7 +188,10 @@ class AstEavesdropper(ast.NodeTransformer):
                         attr_or_subscript=attr_or_sub,
                         ctx=fast.Str(node.ctx.__class__.__name__),
                         call_context=fast.NameConstant(call_context),
-                        top_level_node_id=self._get_copy_id_ast(self._top_level_node_for_symbol)
+                        top_level_node_id=self._get_copy_id_ast(self._top_level_node_for_symbol),
+                        subscript_name=(
+                            fast.NameConstant(subscript_name) if subscript_name is None else fast.Str(subscript_name)
+                        ),
                     ) + extra_args
                 )
         # end fast.location_of(node.value)
@@ -275,8 +294,35 @@ class AstEavesdropper(ast.NodeTransformer):
             )
         return node
 
-    def visit_literal(self, node: Union[ast.Dict, ast.List, ast.Tuple], should_inner_visit=True):
-        # TODO: what about set literals?
+    def visit_Lambda(self, node: ast.Lambda):
+        assert isinstance(getattr(node, 'ctx', ast.Load()), ast.Load)
+        visited = self.generic_visit(node)
+        with fast.location_of(node):
+            subscripted_node = self._make_tuple_event_for(
+                visited, TraceEvent.before_lambda, orig_node_id=id(node)
+            )
+            return fast.Call(
+                func=self._emitter_ast(),
+                args=[TraceEvent.after_lambda.to_ast(), self._get_copy_id_ast(node)],
+                keywords=fast.kwargs(ret=subscripted_node),
+            )
+
+    @staticmethod
+    def _ast_container_to_literal_trace_evt(
+        node: Union[ast.Dict, ast.List, ast.Set, ast.Tuple], before: bool
+    ) -> TraceEvent:
+        if isinstance(node, ast.Dict):
+            return TraceEvent.before_dict_literal if before else TraceEvent.after_dict_literal
+        elif isinstance(node, ast.List):
+            return TraceEvent.before_list_literal if before else TraceEvent.after_list_literal
+        elif isinstance(node, ast.Set):
+            return TraceEvent.before_set_literal if before else TraceEvent.after_set_literal
+        elif isinstance(node, ast.Tuple):
+            return TraceEvent.before_tuple_literal if before else TraceEvent.after_tuple_literal
+        else:
+            raise TypeError('invalid ast node: %s', ast.dump(node))
+
+    def visit_literal(self, node: Union[ast.Dict, ast.List, ast.Set, ast.Tuple], should_inner_visit=True):
         maybe_visited: ast.AST = node
         if should_inner_visit:
             maybe_visited = self.generic_visit(node)
@@ -284,21 +330,38 @@ class AstEavesdropper(ast.NodeTransformer):
             return maybe_visited
         with fast.location_of(node):
             subscripted_node = self._make_tuple_event_for(
-                maybe_visited, TraceEvent.before_literal, orig_node_id=id(node)
+                maybe_visited, self._ast_container_to_literal_trace_evt(node, before=True), orig_node_id=id(node)
             )
             return fast.Call(
                 func=self._emitter_ast(),
-                args=[TraceEvent.after_literal.to_ast(), self._get_copy_id_ast(node)],
+                args=[
+                    self._ast_container_to_literal_trace_evt(node, before=False).to_ast(),
+                    self._get_copy_id_ast(node)
+                ],
                 keywords=fast.kwargs(ret=subscripted_node),
             )
 
-    def visit_Tuple(self, node: ast.Tuple):
-        return self.visit_List_or_Tuple(node)
-
     def visit_List(self, node: ast.List):
-        return self.visit_List_or_Tuple(node)
+        return self.visit_List_or_Set_or_Tuple(node)
 
-    def visit_List_or_Tuple(self, node: Union[ast.List, ast.Tuple]):
+    def visit_Set(self, node: ast.Set):
+        return self.visit_List_or_Set_or_Tuple(node)
+
+    def visit_Tuple(self, node: ast.Tuple):
+        return self.visit_List_or_Set_or_Tuple(node)
+
+    @staticmethod
+    def _ast_container_to_elt_trace_evt(node: Union[ast.List, ast.Set, ast.Tuple]) -> TraceEvent:
+        if isinstance(node, ast.List):
+            return TraceEvent.list_elt
+        elif isinstance(node, ast.Set):
+            return TraceEvent.set_elt
+        elif isinstance(node, ast.Tuple):
+            return TraceEvent.tuple_elt
+        else:
+            raise TypeError('invalid ast node: %s', ast.dump(node))
+
+    def visit_List_or_Set_or_Tuple(self, node: Union[ast.List, ast.Set, ast.Tuple]):
         traced_elts: List[ast.expr] = []
         is_load = isinstance(getattr(node, 'ctx', ast.Load()), ast.Load)
         saw_starred = False
@@ -315,7 +378,7 @@ class AstEavesdropper(ast.NodeTransformer):
                 traced_elts.append(fast.Call(
                     func=self._emitter_ast(),
                     args=[
-                        TraceEvent.list_elt.to_ast() if isinstance(node, ast.List) else TraceEvent.tuple_elt.to_ast(),
+                        self._ast_container_to_elt_trace_evt(node).to_ast(),
                         self._get_copy_id_ast(elt),
                     ],
                     keywords=fast.kwargs(
@@ -364,6 +427,8 @@ class AstEavesdropper(ast.NodeTransformer):
         return self.visit_literal(node, should_inner_visit=False)
 
     def visit_Return(self, node: ast.Return):
+        if node.value is None:
+            return node
         with fast.location_of(node):
             node.value = fast.Call(
                 func=self._emitter_ast(),
