@@ -140,19 +140,7 @@ class ResolveRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitList
         self.symbols.extend(tracer().resolve_loaded_symbols(node))
 
     def visit_Lambda(self, node):
-        with self._push_symbols():
-            self.visit(node.body)
-            self.visit(node.args)
-            to_add = set(self.symbols)
-        # remove node.arguments
-        with self._push_symbols():
-            self.visit(node.args.args)
-            self.visit(node.args.vararg)
-            self.visit(node.args.kwonlyargs)
-            self.visit(node.args.kwarg)
-            discard_set = set(self.symbols)
-        # throw away anything appearing in lambda body that isn't bound
-        self.symbols.extend(to_add - discard_set)
+        self.symbols.extend(tracer().resolve_loaded_symbols(node))
 
     def visit_GeneratorExp(self, node):
         self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)
@@ -216,6 +204,11 @@ class ResolveRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitList
     def visit_withitem(self, node: ast.withitem):
         self.visit(node.context_expr)
 
+    def visit_ExceptHandler(self, node: ast.ExceptHandler):
+        # important: this needs to skip the body
+        if node.type is not None:
+            self.symbols.extend(tracer().resolve_loaded_symbols(node.type))
+
     def visit_Import(self, node: ast.Import):
         pass
 
@@ -229,9 +222,8 @@ def update_usage_info(symbols: Union[Optional[DataSymbol], Set[Optional[DataSymb
         if sym is None:
             continue
         sym.last_used_cell_num = cell_counter
-        if sym.defined_cell_num < cell_counter:
-            logger.info('sym `%s` used in cell %d last updated in cell %d', sym, cell_counter, sym.defined_cell_num)
-            sym.version_by_used_timestamp[cell_counter] = sym.defined_cell_num
+        logger.info('sym `%s` used in cell %d last updated in cell %d', sym, cell_counter, sym.timestamp)
+        sym.timestamp_by_used_time[cell_counter] = sym.timestamp
 
 
 def resolve_rval_symbols(node: Union[str, ast.AST], should_update_usage_info: bool = True) -> Set[DataSymbol]:
