@@ -8,6 +8,8 @@ EMIT_EVENT = '_X5ix_NBSAFETY_EVT_EMIT'
 
 
 class TraceEvent(Enum):
+    init_cell = 'init_cell'
+
     before_stmt = 'before_stmt'
     after_stmt = 'after_stmt'
 
@@ -17,17 +19,28 @@ class TraceEvent(Enum):
     before_complex_symbol = 'before_complex_symbol'
     after_complex_symbol = 'after_complex_symbol'
 
+    before_lambda = 'before_lambda'
+    after_lambda = 'after_lambda'
+
     before_call = 'before_call'
     after_call = 'after_call'
     argument = 'argument'
     before_return = 'before_return'
     after_return = 'after_return'
 
-    before_literal = 'before_literal'
-    after_literal = 'after_literal'
+    before_dict_literal = 'before_dict_literal'
+    after_dict_literal = 'after_dict_literal'
+    before_list_literal = 'before_list_literal'
+    after_list_literal = 'after_list_literal'
+    before_set_literal = 'before_set_literal'
+    after_set_literal = 'after_set_literal'
+    before_tuple_literal = 'before_tuple_literal'
+    after_tuple_literal = 'after_tuple_literal'
+
     dict_key = 'dict_key'
     dict_value = 'dict_value'
     list_elt = 'list_elt'
+    set_elt = 'set_elt'
     tuple_elt = 'tuple_elt'
 
     before_assign_rhs = 'before_assign_rhs'
