@@ -3,18 +3,20 @@ import ast
 import builtins
 from collections import defaultdict
 from contextlib import contextmanager
+import functools
 import logging
 import sys
 from typing import cast, TYPE_CHECKING
 
 import astunparse
+from IPython import get_ipython
 
-from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
-from nbsafety.data_model.scope import Scope, NamespaceScope
 from nbsafety import singletons
+from nbsafety.data_model.data_symbol import DataSymbol
+from nbsafety.data_model.scope import Scope, NamespaceScope
 from nbsafety.run_mode import SafetyRunMode
 from nbsafety.singletons import nbs
-from nbsafety.tracing.mutation_event import MutationEvent
+from nbsafety.tracing.mutation_event import ArgMutate, ListAppend, ListExtend, ListInsert, StandardMutation
 from nbsafety.tracing.symbol_resolver import resolve_rval_symbols
 from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
 from nbsafety.tracing.trace_stack import TraceStack
@@ -24,14 +26,16 @@ from nbsafety.tracing.utils import match_container_obj_or_namespace_with_literal
 if TYPE_CHECKING:
     from typing import Any, Callable, DefaultDict, Dict, List, Optional, Set, Tuple, Type, Union
     from types import FrameType
-    AttrSubVal = Union[str, int]
+    from nbsafety.tracing.mutation_event import MutationEvent
+    from nbsafety.types import SupportedIndexType
+    AttrSubVal = SupportedIndexType
     NodeId = int
     ObjId = int
     MutationCandidate = Tuple[Tuple[int, ObjId, Optional[str]], MutationEvent, Set[DataSymbol], List[Any]]
     Mutation = Tuple[int, MutationEvent, Set[DataSymbol], List[Any]]
     SavedStoreData = Tuple[NamespaceScope, Any, AttrSubVal, bool]
     SavedDelData = Tuple[NamespaceScope, AttrSubVal, bool]
-    SavedComplexSymbolLoadData = Tuple[NamespaceScope, Tuple[AttrSubVal, bool]]
+    SavedComplexSymbolLoadData = Tuple[Tuple[NamespaceScope, Any], Tuple[AttrSubVal, bool]]
 
 
 logger = logging.getLogger(__name__)
@@ -53,7 +57,6 @@ ARG_MUTATION_EXCEPTED_MODULES = {
 }
 
 
-
 def _finish_tracing_reset():
     # do nothing; we just want to trigger the newly reenabled tracer with a 'call' event
     pass
@@ -63,7 +66,7 @@ class BaseTraceManager(singletons.TraceManager):
 
     _MANAGER_CLASS_REGISTERED = False
     EVENT_HANDLERS_PENDING_REGISTRATION: DefaultDict[TraceEvent, List[Callable[..., Any]]] = defaultdict(list)
-    EVENT_HANDLERS_BY_CLASS: Dict[Type['BaseTraceManager'], DefaultDict[TraceEvent, List[Callable[..., Any]]]] = {}
+    EVENT_HANDLERS_BY_CLASS: Dict[Type[BaseTraceManager], DefaultDict[TraceEvent, List[Callable[..., Any]]]] = {}
 
     EVENT_LOGGER = logging.getLogger('events')
     EVENT_LOGGER.setLevel(logging.WARNING)
@@ -77,6 +80,8 @@ class BaseTraceManager(singletons.TraceManager):
         self._event_handlers = self.EVENT_HANDLERS_BY_CLASS[self.__class__]
         self.tracing_enabled = False
         self.tracing_reset_pending = False
+        self.sys_tracer = self._sys_tracer
+        self.existing_tracer = None
 
     def _emit_event(self, evt: Union[TraceEvent, str], node_id: int, **kwargs: Any):
         event = TraceEvent(evt) if isinstance(evt, str) else evt
@@ -89,7 +94,7 @@ class BaseTraceManager(singletons.TraceManager):
                 if SafetyRunMode.get() == SafetyRunMode.DEVELOP:
                     raise exc
                 else:
-                    logger.error('Exception occurred: %s' % exc)
+                    logger.error('Exception occurred: %s', str(exc))
                 new_ret = None
             if new_ret is not None:
                 kwargs['ret'] = new_ret
@@ -98,23 +103,64 @@ class BaseTraceManager(singletons.TraceManager):
     def _make_stack(self):
         return TraceStack(self)
 
-    def _enable_tracing(self):
-        assert not self.tracing_enabled
+    def _make_composed_tracer(self, existing_tracer):  # pragma: no cover
+
+        @functools.wraps(self._sys_tracer)
+        def _composed_tracer(frame: FrameType, evt: str, arg: Any, **kwargs):
+            existing_ret = existing_tracer(frame, evt, arg, **kwargs)
+            if not self.tracing_enabled:
+                return existing_ret
+            my_ret = self._sys_tracer(frame, evt, arg, **kwargs)
+            if my_ret is None and evt == 'call':
+                return existing_ret
+            else:
+                return my_ret
+        return _composed_tracer
+
+    def _settrace_patch(self, trace_func):  # pragma: no cover
+        # called by third-party tracers
+        self.existing_tracer = trace_func
+        if self.tracing_enabled:
+            if trace_func is None:
+                self._disable_tracing()
+            self._enable_tracing(check_disabled=False, existing_tracer=trace_func)
+        else:
+            nbs().settrace(trace_func)
+
+    def _enable_tracing(self, check_disabled=True, existing_tracer=None):
+        if check_disabled:
+            assert not self.tracing_enabled
         self.tracing_enabled = True
-        sys.settrace(self._sys_tracer)
+        self.existing_tracer = existing_tracer or sys.gettrace()
+        if self.existing_tracer is None:
+            self.sys_tracer = self._sys_tracer
+        else:
+            self.sys_tracer = self._make_composed_tracer(self.existing_tracer)
+        nbs().settrace(self.sys_tracer)
 
     def _disable_tracing(self, check_enabled=True):
         if check_enabled:
             assert self.tracing_enabled
+            assert sys.gettrace() is self.sys_tracer
         self.tracing_enabled = False
-        sys.settrace(None)
+        nbs().settrace(self.existing_tracer)
+
+    @contextmanager
+    def _patch_sys_settrace(self):
+        original_settrace = sys.settrace
+        try:
+            sys.settrace = self._settrace_patch
+            yield
+        finally:
+            sys.settrace = original_settrace
 
     @contextmanager
     def tracing_context(self):
+        setattr(builtins, EMIT_EVENT, self._emit_event)
         try:
-            setattr(builtins, EMIT_EVENT, self._emit_event)
-            self._enable_tracing()
-            yield
+            with self._patch_sys_settrace():
+                self._enable_tracing()
+                yield
         finally:
             delattr(builtins, EMIT_EVENT)
             self._disable_tracing(check_enabled=False)
@@ -127,8 +173,7 @@ class BaseTraceManager(singletons.TraceManager):
             assert evt == 'call', 'expected call; got event %s' % evt
             self._attempt_to_reenable_tracing(frame)
             return None
-        # notebook cells have filenames that appear as '<ipython-input...>'
-        if evt == 'line' or not frame.f_code.co_filename.startswith('<ipython-input'):
+        if evt == 'line' or not nbs().is_cell_file(frame.f_code.co_filename):
             return None
 
         return self._emit_event(evt, 0, _frame=frame, ret=arg)
@@ -183,7 +228,7 @@ class TraceManager(BaseTraceManager):
             with self.call_stack.needing_manual_initialization():
                 self.cur_frame_original_scope: Scope = nbs().global_scope
                 self.active_scope: Scope = nbs().global_scope
-                self.inside_lambda = False
+                self.inside_anonymous_call = False
 
             self.lexical_call_stack: TraceStack = self._make_stack()
             with self.lexical_call_stack.register_stack_state():
@@ -214,15 +259,13 @@ class TraceManager(BaseTraceManager):
         self.node_id_to_saved_dict_key.clear()
 
     def _handle_call_transition(self, trace_stmt: TraceStatement):
-        node_before_call = nbs().ast_node_by_id[self.prev_node_id_in_cur_frame]
-        if hasattr(node_before_call, 'ctx') and isinstance(getattr(node_before_call, 'ctx'), ast.Del):
-            self.node_id_to_saved_del_data.pop(self.prev_node_id_in_cur_frame, None)
-        self.node_id_to_saved_del_data.clear()  # ensures we only handle del's and not delitem's
+        # ensures we only handle del's and not delitem's
+        self.node_id_to_saved_del_data.clear()
         new_scope = trace_stmt.get_post_call_scope()
         with self.call_stack.push():
             # TODO: figure out a better way to determine if we're inside a lambda
             #  could this one lead to a false negative if a lambda is in the default of a function def kwarg?
-            self.inside_lambda = not isinstance(
+            self.inside_anonymous_call = not isinstance(
                 trace_stmt.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
             )
             self.cur_frame_original_scope = new_scope
@@ -243,7 +286,7 @@ class TraceManager(BaseTraceManager):
 
     def _handle_return_transition(self, trace_stmt: TraceStatement, ret: Any):
         try:
-            inside_lambda = self.inside_lambda
+            inside_anonymous_call = self.inside_anonymous_call
             return_to_stmt: TraceStatement = self.call_stack.get_field('prev_trace_stmt_in_cur_frame')
             assert return_to_stmt is not None
             if self.prev_event != TraceEvent.exception:
@@ -251,10 +294,17 @@ class TraceManager(BaseTraceManager):
                 # no need to track dependencies in this case
                 if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
                     return_to_stmt.class_scope = cast(NamespaceScope, self.cur_frame_original_scope)
-                elif isinstance(trace_stmt.stmt_node, ast.Return) or inside_lambda:
+                elif isinstance(trace_stmt.stmt_node, ast.Return) or inside_anonymous_call:
                     if not trace_stmt.lambda_call_point_deps_done_once:
                         trace_stmt.lambda_call_point_deps_done_once = True
-                        rvals = resolve_rval_symbols(trace_stmt.stmt_node)
+                        maybe_lambda_sym = nbs().statement_to_func_cell.get(id(trace_stmt.stmt_node), None)
+                        maybe_lambda_node = None
+                        if maybe_lambda_sym is not None:
+                            maybe_lambda_node = maybe_lambda_sym.stmt_node
+                        if inside_anonymous_call and maybe_lambda_node is not None and isinstance(maybe_lambda_node, ast.Lambda):
+                            rvals = resolve_rval_symbols(maybe_lambda_node.body)
+                        else:
+                            rvals = resolve_rval_symbols(trace_stmt.stmt_node)
                         dsym_to_attach = None
                         if len(rvals) == 1:
                             dsym_to_attach = next(iter(rvals))
@@ -275,7 +325,7 @@ class TraceManager(BaseTraceManager):
                                 # logger.error("prev seen outer: %s", ast.dump(nbs().ast_node_by_id[call_node_id]))
                                 total_args = len(call_node.args) + len(call_node.keywords)
                                 num_args_seen = self.call_stack.get_field('num_args_seen')
-                                logger.info("num args seen: %d", num_args_seen)
+                                logger.warning("num args seen: %d", num_args_seen)
                                 if total_args == num_args_seen:
                                     return_to_node_id = call_node_id
                                 else:
@@ -317,36 +367,34 @@ class TraceManager(BaseTraceManager):
             ref = id(ref)
         return ref
 
-    def resolve_store_or_del_data_for_target(
-        self, target: Union[str, int, ast.AST], frame: FrameType, ctx: Union[ast.Del, ast.Store] = ast.Store()
-    ) -> Tuple[Scope, Union[str, int], Any, bool]:
+    def resolve_store_data_for_target(
+        self, target: Union[str, int, ast.AST], frame: FrameType
+    ) -> Tuple[Scope, AttrSubVal, Any, bool]:
         target = self._partial_resolve_ref(target)
         if isinstance(target, str):
-            if isinstance(ctx, ast.Store):
-                obj = frame.f_locals[target]
-            else:
-                obj = None
+            obj = frame.f_locals[target]
             return self.cur_frame_original_scope, target, obj, False
+        (
+            scope, obj, attr_or_sub, is_subscript
+        ) = self.node_id_to_saved_store_data[target]
+        attr_or_sub_obj = nbs().retrieve_namespace_attr_or_sub(obj, attr_or_sub, is_subscript)
+        if attr_or_sub_obj is None:
+            scope_to_use = scope
         else:
-            if isinstance(ctx, ast.Store):
-                (
-                    scope, obj, attr_or_sub, is_subscript
-                ) = self.node_id_to_saved_store_data[target]
-                attr_or_sub_obj = nbs().retrieve_namespace_attr_or_sub(obj, attr_or_sub, is_subscript)
-            else:
-                assert isinstance(ctx, ast.Del)
-                (
-                    scope, attr_or_sub, is_subscript
-                ) = self.node_id_to_saved_del_data[target]
-                attr_or_sub_obj = None
-            if attr_or_sub_obj is None:
-                scope_to_use = scope
-            else:
-                scope_to_use = scope.get_earliest_ancestor_containing(id(attr_or_sub_obj), is_subscript)
-            if scope_to_use is None:
-                # Nobody before `scope` has it, so we'll insert it at this level
-                scope_to_use = scope
-            return scope_to_use, attr_or_sub, attr_or_sub_obj, is_subscript
+            scope_to_use = scope.get_earliest_ancestor_containing(id(attr_or_sub_obj), is_subscript)
+        if scope_to_use is None:
+            # Nobody before `scope` has it, so we'll insert it at this level
+            scope_to_use = scope
+        return scope_to_use, attr_or_sub, attr_or_sub_obj, is_subscript
+
+    def resolve_del_data_for_target(self, target: Union[str, int, ast.AST]) -> Tuple[Scope, AttrSubVal, Any, bool]:
+        target = self._partial_resolve_ref(target)
+        if isinstance(target, str):
+            return self.cur_frame_original_scope, target, None, False
+        (
+            scope, attr_or_sub, is_subscript
+        ) = self.node_id_to_saved_del_data[target]
+        return scope, attr_or_sub, None, is_subscript
 
     def resolve_loaded_symbols(self, symbol_ref: Union[str, int, ast.AST, DataSymbol]) -> List[DataSymbol]:
         if isinstance(symbol_ref, DataSymbol):
@@ -372,7 +420,7 @@ class TraceManager(BaseTraceManager):
             return ns
         class_scope = nbs().namespaces.get(id(obj.__class__), None)
         if class_scope is not None:
-            # logger.info(
+            # logger.warning(
             #     'found class scope %s containing %s',
             #     class_scope, list(class_scope.all_data_symbols_this_indentation())
             # )
@@ -401,27 +449,28 @@ class TraceManager(BaseTraceManager):
     def _clear_info_and_maybe_lookup_or_create_complex_symbol(self, obj_attr_or_sub) -> Optional[DataSymbol]:
         if self.saved_complex_symbol_load_data is None:
             return None
-        scope, (attr_or_subscript, is_subscript) = self.saved_complex_symbol_load_data
+        (scope, obj), (attr_or_subscript, is_subscript) = self.saved_complex_symbol_load_data
         self.saved_complex_symbol_load_data = None
         data_sym = scope.lookup_data_symbol_by_name_this_indentation(
             attr_or_subscript, is_subscript=is_subscript, skip_cloned_lookup=True,
         )
-        logger.info("found sym %s in scope %s", data_sym, scope)
+        logger.warning("found sym %s in scope %s", data_sym, scope)
         if data_sym is None:
             parent = scope.lookup_data_symbol_by_name_this_indentation(
                 attr_or_subscript, is_subscript, skip_cloned_lookup=False,
             )
             parents = set() if parent is None else {parent}
+            is_default_dict = isinstance(obj, defaultdict)
             data_sym = scope.upsert_data_symbol_for_name(
                 attr_or_subscript,
                 obj_attr_or_sub,
                 parents,
                 self.prev_trace_stmt_in_cur_frame.stmt_node,
                 is_subscript=is_subscript,
-                propagate=False,
-                implicit=True,
+                propagate=is_default_dict,
+                implicit=not is_default_dict,
             )
-            logger.info("create implicit sym %s", data_sym)
+            logger.warning("create implicit sym %s", data_sym)
         elif data_sym.obj_id != id(obj_attr_or_sub):
             data_sym.update_obj_ref(obj_attr_or_sub)
         return data_sym
@@ -442,6 +491,10 @@ class TraceManager(BaseTraceManager):
         self.prev_node_id_in_cur_frame = node_id
         self.prev_node_id_in_cur_frame_lexical = node_id
 
+    @register_handler(TraceEvent.init_cell)
+    def init_cell(self, _obj, _node_id, frame: FrameType, _event, cell_id: Union[str, int], **__):
+        nbs().set_name_to_cell_num_mapping(frame)
+
     @register_handler(TraceEvent.after_assign_rhs)
     def after_assign_rhs(self, obj: Any, *_, **__):
         if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
@@ -456,7 +509,7 @@ class TraceManager(BaseTraceManager):
         _frame_: FrameType,
         event: TraceEvent,
         *_,
-        attr_or_subscript,
+        attr_or_subscript: AttrSubVal,
         ctx: str,
         call_context: bool,
         top_level_node_id: NodeId,
@@ -468,11 +521,20 @@ class TraceManager(BaseTraceManager):
         if isinstance(nbs().ast_node_by_id[node_id], ast.Call):
             # clear the callpoint dependency
             self.node_id_to_loaded_symbols.pop(node_id, None)
-        if obj is None:
+        if obj is None or obj is get_ipython():
             return
-        logger.info('%s attrsub %s of obj %s', ctx, attr_or_subscript, obj)
+        logger.warning('%s attrsub %s of obj %s', ctx, attr_or_subscript, obj)
         sym_for_obj = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
+        
+        # Resolve symbol if necessary
+        if sym_for_obj is None and obj_name is not None:
+            sym_for_obj = self.active_scope.lookup_data_symbol_by_name_this_indentation(obj_name)
+
+        if sym_for_obj is not None and sym_for_obj.timestamp < nbs().cell_counter():
+            sym_for_obj.timestamp_by_used_time[nbs().cell_counter()] = sym_for_obj.timestamp_excluding_ns_descendents
+        
         is_subscript = (event == TraceEvent.subscript)
+
         obj_id = id(obj)
         if self.top_level_node_id_for_chain is None:
             self.top_level_node_id_for_chain = top_level_node_id
@@ -486,8 +548,9 @@ class TraceManager(BaseTraceManager):
 
         scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
         self.active_scope = scope
+
         if ctx in ('Store', 'AugStore'):
-            logger.info(
+            logger.warning(
                 "save store data for node id %d: %s, %s, %s, %s",
                 top_level_node_id, scope, obj, attr_or_subscript, is_subscript
             )
@@ -495,13 +558,18 @@ class TraceManager(BaseTraceManager):
             return
         elif ctx == 'Del':
             # logger.error("save del data for node %s", ast.dump(nbs().ast_node_by_id[top_level_node_id]))
-            logger.info("save del data for node id %d", top_level_node_id)
+            logger.warning("save del data for node id %d", top_level_node_id)
             self.node_id_to_saved_del_data[top_level_node_id] = (scope, attr_or_subscript, is_subscript)
             return
         if call_context:
-            mutation_event = MutationEvent.normal
-            if isinstance(obj, list) and attr_or_subscript == 'append':
-                mutation_event = MutationEvent.list_append
+            mutation_event: MutationEvent = StandardMutation()
+            if isinstance(obj, list):
+                if attr_or_subscript == 'append':
+                    mutation_event = ListAppend()
+                elif attr_or_subscript == 'extend':
+                    mutation_event = ListExtend(len(obj))
+                elif attr_or_subscript == 'insert':
+                    mutation_event = ListInsert()
             # save off event counter and obj_id
             # if event counter didn't change when we process the Call retval, and if the
             # retval is None, this is a likely signal that we have a mutation
@@ -521,13 +589,14 @@ class TraceManager(BaseTraceManager):
                         self.prev_trace_stmt_in_cur_frame.stmt_node,
                         is_subscript=is_subscript,
                         is_anonymous=obj_name is None,
+                        propagate=False,
                         implicit=True,
                     )
                 if sym_for_obj is not None:
                     self.sym_for_obj_calling_method = sym_for_obj
         else:
-            logger.info("saved load data: %s, %s, %s", scope, attr_or_subscript, is_subscript)
-            self.saved_complex_symbol_load_data = (scope, (attr_or_subscript, is_subscript))
+            logger.warning("saved load data: %s, %s, %s", scope, attr_or_subscript, is_subscript)
+            self.saved_complex_symbol_load_data = ((scope, obj), (attr_or_subscript, is_subscript))
 
     @register_handler(TraceEvent.after_complex_symbol)
     def after_complex_symbol(self, obj: Any, *_, call_context: bool, ctx: str, **__):
@@ -541,10 +610,15 @@ class TraceManager(BaseTraceManager):
                 return
             loaded_sym = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
             if call_context and len(self.mutation_candidates) > 0:
-                (evt_counter, obj_id, obj_name), mutation_event, recorded_arg_dsyms, recorded_arg_objs = self.mutation_candidates.pop()
+                (
+                    (evt_counter, obj_id, obj_name),
+                    mutation_event,
+                    recorded_arg_dsyms,
+                    recorded_arg_objs,
+                ) = self.mutation_candidates.pop()
                 if evt_counter == self.trace_event_counter:
-                    if obj is None:
-                        if mutation_event == MutationEvent.normal:
+                    if obj is None or id(obj) == obj_id:
+                        if isinstance(mutation_event, StandardMutation):
                             try:
                                 top_level_sym = nbs().get_first_full_symbol(self.first_obj_id_in_chain)
                                 if top_level_sym.is_import and top_level_sym.name not in ARG_MUTATION_EXCEPTED_MODULES:
@@ -552,9 +626,11 @@ class TraceManager(BaseTraceManager):
                                     #  i.e. allow-list for arg mutations, starting with np.random.seed?
                                     if len(recorded_arg_dsyms) > 0:
                                         # only make this an arg mutation event if it looks like there's an arg to mutate
-                                        mutation_event = MutationEvent.arg_mutate
+                                        mutation_event = ArgMutate()
                             except:
                                 pass
+                        elif isinstance(mutation_event, ListInsert):
+                            mutation_event.insert_pos = recorded_arg_objs[0]
                         self.mutations.append((obj_id, mutation_event, recorded_arg_dsyms, recorded_arg_objs))
                     else:
                         if self.sym_for_obj_calling_method is not None:
@@ -575,8 +651,6 @@ class TraceManager(BaseTraceManager):
             return
         self.num_args_seen += 1
         arg_node = nbs().ast_node_by_id.get(arg_node_id, None)
-        if not isinstance(arg_node, (ast.Attribute, ast.Subscript, ast.Call, ast.Name, ast.Num, ast.Constant, ast.Str)):
-            return
         if len(self.mutation_candidates) == 0:
             return
 
@@ -584,13 +658,10 @@ class TraceManager(BaseTraceManager):
             assert self.active_scope is self.cur_frame_original_scope
             arg_dsym = self.active_scope.lookup_data_symbol_by_name(arg_node.id)
             if arg_dsym is None:
-                arg_dsym = self.active_scope.upsert_data_symbol_for_name(
+                self.active_scope.upsert_data_symbol_for_name(
                     arg_node.id, arg_obj, set(), self.prev_trace_stmt_in_cur_frame.stmt_node, implicit=True
                 )
-            arg_dsyms = [arg_dsym]
-        else:
-            arg_dsyms = self.node_id_to_loaded_symbols.get(arg_node_id, [])
-        self.mutation_candidates[-1][-2].update(arg_dsyms)
+        self.mutation_candidates[-1][-2].update(resolve_rval_symbols(arg_node))
         self.mutation_candidates[-1][-1].append(arg_obj)
 
     @register_handler(TraceEvent.before_call)
@@ -609,7 +680,8 @@ class TraceManager(BaseTraceManager):
         # that will happen in the 'after chain' handler
         self.lexical_call_stack.pop()
 
-    @register_handler(TraceEvent.before_literal)
+    # Note: we don't trace set literals
+    @register_handler((TraceEvent.before_dict_literal, TraceEvent.before_list_literal, TraceEvent.before_tuple_literal))
     def before_literal(self, *_, **__):
         if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
             return
@@ -617,13 +689,13 @@ class TraceManager(BaseTraceManager):
         with self.lexical_literal_stack.push():
             self.active_literal_scope = NamespaceScope(None, NamespaceScope.ANONYMOUS, parent_scope)
 
-    @register_handler(TraceEvent.after_literal)
-    def after_literal(self, literal: Any, node_id: NodeId, *_, **__):
+    @register_handler((TraceEvent.after_dict_literal, TraceEvent.after_list_literal, TraceEvent.after_tuple_literal))
+    def after_literal(self, literal: Union[dict, list, tuple], node_id: NodeId, *_, **__):
         if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
             return literal
         try:
             self.active_literal_scope.update_obj_ref(literal)
-            logger.info("create literal scope %s", self.active_literal_scope)
+            logger.warning("create literal scope %s", self.active_literal_scope)
             starred_idx = -1
             starred_namespace = None
             for (i, inner_obj), (inner_key_node, inner_val_node) in match_container_obj_or_namespace_with_literal_nodes(
@@ -646,7 +718,7 @@ class TraceManager(BaseTraceManager):
                         inner_symbols.update(resolve_rval_symbols(inner_key_node))
                 self.node_id_to_loaded_symbols.pop(id(inner_val_node), None)
                 inner_symbols.discard(None)
-                if isinstance(i, (int, str)):
+                if isinstance(i, (int, str)):  # TODO: perform more general check for SupportedIndexType
                     self.active_literal_scope.upsert_data_symbol_for_name(
                         i, inner_obj, inner_symbols, self.prev_trace_stmt_in_cur_frame.stmt_node, is_subscript=True
                     )
@@ -699,6 +771,26 @@ class TraceManager(BaseTraceManager):
             scope.scope_name = str(index)
         return obj
 
+    @register_handler(TraceEvent.after_lambda)
+    def after_lambda(self, obj: Any, lambda_node_id: int, frame: FrameType, *_, **__):
+        sym_deps = []
+        node = nbs().ast_node_by_id[lambda_node_id]
+        for kw_default in node.args.defaults:  # type: ignore
+            sym_deps.extend(self.resolve_loaded_symbols(kw_default))
+        sym = self.active_scope.upsert_data_symbol_for_name(
+            '<lambda_sym_%d>' % id(obj),
+            obj,
+            sym_deps,
+            self.prev_trace_stmt_in_cur_frame.stmt_node,
+            is_function_def=True,
+            propagate=False,
+        )
+        # FIXME: this is super brittle. We're passing in a stmt node to update the mapping from
+        #  stmt_node to function symbol, but simultaneously forcing the lambda symbol to hold
+        #  a reference to the lambda in order to help with symbol resolution later
+        sym.stmt_node = node
+        self.node_id_to_loaded_symbols[lambda_node_id].append(sym)
+
     @register_handler(TraceEvent.after_stmt)
     def after_stmt(self, ret_expr: Any, stmt_id: int, frame: FrameType, *_, **__):
         if stmt_id in self.seen_stmts:
@@ -737,7 +829,7 @@ class TraceManager(BaseTraceManager):
         self.tracing_reset_pending = False
         call_depth = 0
         while frame is not None:
-            if frame.f_code.co_filename.startswith('<ipython-input'):
+            if nbs().is_cell_file(frame.f_code.co_filename):
                 call_depth += 1
             frame = frame.f_back
         if nbs().is_develop:
@@ -768,16 +860,15 @@ class TraceManager(BaseTraceManager):
         **__
     ):
         # right now, this should only be enabled for notebook code
-        assert frame.f_code.co_filename.startswith('<ipython-input')
+        assert nbs().is_cell_file(frame.f_code.co_filename), 'got %s' % frame.f_code.co_filename
         assert self.tracing_enabled or event == TraceEvent.after_stmt
-        nbs().maybe_set_name_to_cell_num_mapping(frame)
 
         # IPython quirk -- every line in outer scope apparently wrapped in lambda
         # We want to skip the outer 'call' and 'return' for these
         if event == TraceEvent.call:
             self.call_depth += 1
             if self.call_depth == 1:
-                return self._sys_tracer
+                return self.sys_tracer
 
         if event == TraceEvent.return_:
             self.call_depth -= 1
@@ -785,17 +876,30 @@ class TraceManager(BaseTraceManager):
                 return
 
         cell_num, lineno = nbs().get_position(frame)
+        if cell_num is None:
+            return None
 
         if event == TraceEvent.after_stmt:
             assert stmt_node is not None
         else:
             try:
                 stmt_node = nbs().statement_cache[cell_num][lineno]
+                if event == TraceEvent.call and not isinstance(stmt_node, (ast.AsyncFunctionDef, ast.FunctionDef)):
+                    # TODO: this is bad and I should feel bad. Need a better way to figure out which
+                    #  stmt is executing than by using line numbers.
+                    parent_node = nbs().parent_node_by_id.get(id(stmt_node), None)
+                    logger.error("node %s parent %s", ast.dump(stmt_node), None if parent_node is None else ast.dump(parent_node))
+                    if (
+                        parent_node is not None
+                        and getattr(parent_node, 'lineno', None) == lineno
+                        and isinstance(parent_node, (ast.AsyncFunctionDef, ast.FunctionDef))
+                    ):
+                        stmt_node = parent_node
             except KeyError as e:
                 self.EVENT_LOGGER.warning("got key error for stmt node in cell %d, line %d", cell_num, lineno)
                 if nbs().is_develop:
                     raise e
-                return self._sys_tracer
+                return self.sys_tracer
 
         trace_stmt = self.traced_statements.get(id(stmt_node), None)
         if trace_stmt is None:
@@ -814,7 +918,7 @@ class TraceManager(BaseTraceManager):
                 return None
             trace_stmt.node_id_for_last_call = self.prev_node_id_in_cur_frame
         self.state_transition_hook(event, trace_stmt, ret_obj)
-        return self._sys_tracer
+        return self.sys_tracer
 
 
 assert not BaseTraceManager._MANAGER_CLASS_REGISTERED
