@@ -1,10 +1,14 @@
 # -*- coding: future_annotations -*-
 from contextlib import contextmanager
 import logging
+from typing import TYPE_CHECKING
 
 from nbsafety.safety import NotebookSafetySettings
 from nbsafety.singletons import nbs
-from .utils import make_safety_fixture, skipif_known_failing
+from test.utils import make_safety_fixture, skipif_known_failing
+
+if TYPE_CHECKING:
+    from typing import Dict
 
 logging.basicConfig(level=logging.ERROR)
 
@@ -36,21 +40,24 @@ def run_cell(cell, cell_id=None, **kwargs):
     run_cell_(cell, **kwargs)
 
 
+def run_all_cells(cells: Dict[int, str], **kwargs):
+    for cell_id in sorted(cells.keys()):
+        run_cell(cells[cell_id], cell_id=cell_id, **kwargs)
+
+
 def test_simple():
     cells = {
         0: 'x = 0',
         1: 'y = x + 1',
-        2: 'x = 42',
-        3: 'logging.info(y)',
+        2: 'logging.info(y)',
+        3: 'x = 42',
     }
-    run_cell(cells[0])
-    run_cell(cells[1])
-    run_cell(cells[2])
+    run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
-    assert response['stale_cells'] == [3]
-    assert response['fresh_cells'] == []
-    assert response['stale_links'] == {3: [1]}
-    assert response['refresher_links'] == {1: [3]}
+    assert response['stale_cells'] == [2]
+    assert response['fresh_cells'] == [1]
+    assert response['stale_links'] == {2: [1]}
+    assert response['refresher_links'] == {1: [2]}
 
 
 def test_refresh_after_exception_fixed():
@@ -73,10 +80,7 @@ def test_refresh_after_val_changed():
         2: 'logging.info(y)',
         3: 'y = 42',
     }
-    run_cell(cells[0], 0)
-    run_cell(cells[1], 1)
-    run_cell(cells[2], 2)
-    run_cell(cells[3], 3)
+    run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['fresh_cells'] == [2]
 
@@ -89,8 +93,7 @@ def test_inner_mutation_considered_fresh():
         3: 'logging.info(lst)',
         4: 'lst_0.append(42)',
     }
-    for idx, cell in cells.items():
-        run_cell(cell, idx)
+    run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [2, 3]
@@ -122,25 +125,24 @@ for foo in lst:
         3: 'logging.info(lst)',
     }
 
-    for idx, cell in cells.items():
-        run_cell(cell, idx)
+    run_all_cells(cells)
 
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
-    assert response['fresh_cells'] == []
+    assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
 
     cells[4] = 'x.inc()'
     run_cell(cells[4], 4)
 
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == [], 'got %s' % response['stale_cells']
-    assert response['fresh_cells'] == [2, 3]
+    assert response['fresh_cells'] == [2, 3], 'got %s' % response['fresh_cells']
 
     cells[5] = 'foo.inc()'
     run_cell(cells[5], 5)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
-    assert response['fresh_cells'] == [2, 3, 4]
+    assert response['fresh_cells'] == [2, 3, 4], 'got %s' % response['fresh_cells']
 
     if force_subscript_symbol_creation:
         cells[6] = 'lst[-1]'
@@ -162,8 +164,7 @@ def test_no_freshness_for_alias_assignment_post_mutation():
         1: 'y = x',
         2: 'x.append(5)',
     }
-    for idx, cell in cells.items():
-        run_cell(cell, idx)
+    run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == []
@@ -174,8 +175,7 @@ def test_fresh_after_import():
         0: 'x = np.random.random(10)',
         1: 'import numpy as np'
     }
-    for idx, cell in cells.items():
-        run_cell(cell, idx, ignore_exceptions=True)
+    run_all_cells(cells, ignore_exceptions=True)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [0]
@@ -191,11 +191,10 @@ def test_external_object_update_propagates_to_stale_namespace_symbols():
         5: 'x = 43',
         6: 'foo = foo.set_x(10)',
     }
-    with override_settings(skip_unsafe_cells=False):
-        for idx, cell in cells.items():
-            run_cell(cell, idx)
+    with override_settings(mark_stale_symbol_usages_unsafe=False):
+        run_all_cells(cells)
         response = nbs().check_and_link_multiple_cells(cells)
-        assert response['stale_cells'] == []
+        assert response['stale_cells'] == [], 'got %s' % response['stale_cells']
         assert response['fresh_cells'] == [2, 4]
 
 
@@ -203,13 +202,12 @@ def test_symbol_on_both_sides_of_assignment():
     cells = {
         0: 'x = 0',
         1: 'y = x + 1',
-        2: 'x = 42',
+        2: 'y += 7',
+        3: 'x = 42',
     }
-    for idx, cell in cells.items():
-        run_cell(cell, idx)
-    cells[3] = 'y += 7'
+    run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
-    assert response['stale_cells'] == [3]
+    assert response['stale_cells'] == [2]
     assert response['fresh_cells'] == [1]
     assert list(response['refresher_links'].keys()) == [1]
 
@@ -221,12 +219,11 @@ def test_updated_namespace_after_subscript_dep_removed():
         2: 'logging.info(d[5])',
         3: 'x = 9',
     }
-    for idx, cell in cells.items():
-        run_cell(cell, idx)
+    run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == [2]
     assert response['fresh_cells'] == [1]
-    cells[1] = 'd = {5: 5}'
+    cells[1] = 'd = {5: 6}'
     run_cell(cells[1], 1)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
@@ -239,3 +236,113 @@ def test_updated_namespace_after_subscript_dep_removed():
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
+
+
+def test_equal_list_update_does_not_induce_fresh_cell():
+    cells = {
+        0: 'x = ["f"] + ["o"] * 10',
+        1: 'y = x + list("bar")',
+        2: 'logging.info(y)',
+        3: 'y = list("".join(y))',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('y = ["f"]', 4)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2, 3]
+
+
+def test_equal_dict_update_does_not_induce_fresh_cell():
+    cells = {
+        0: 'x = {"foo": 42, "bar": 43}',
+        1: 'y = dict(set(x.items()) | set({"baz": 44}.items()))',
+        2: 'logging.info(y)',
+        3: 'y = dict(y.items())',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
+    run_cell('y = {"foo": 99}', 4)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2, 3]
+
+
+def test_list_append():
+    cells = {
+        0: 'lst = [0, 1]',
+        1: 'x = lst[1] + 1',
+        2: 'logging.info(x)',
+        3: 'lst.append(2)',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('lst[1] += 42', 4)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == [2]
+    assert response['fresh_cells'] == [1, 3]
+
+
+def test_list_extend():
+    cells = {
+        0: 'lst = [0, 1]',
+        1: 'x = lst[1] + 1',
+        2: 'logging.info(x)',
+        3: 'lst.extend([2, 3, 4])',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('lst[1] += 42', 4)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == [2]
+    assert response['fresh_cells'] == [1, 3]
+
+
+def test_implicit_subscript_symbol_does_not_bump_ts():
+    cells = {
+        0: 'lst = [] + [0, 1]',
+        1: 'logging.info(lst)',
+        2: 'logging.info(lst[0])',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+
+
+def test_liveness_skipped_for_simple_assignment_involving_aliases():
+    cells = {
+        0: 'lst = [1, 2, 3]',
+        1: 'lst2 = lst',
+        2: 'lst.append(4)',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('lst = [1, 2, 3, 4]', 3)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [1], 'got %s' % response['fresh_cells']
+
+
+def test_incorrect_object_not_used_for_argument_symbols():
+    cells = {
+        0: 'import numpy as np',
+        1: 'arr = np.random.randn(100)',
+        2: 'def f(x): return []',
+        # when obj for `np` was passed to `x`, this created issues
+        3: 'f(np.arange(10))',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
